/*
 * feedbackd - dynamic feedback system for LVS
 * Copyright (C) 2002 Jeremy Kerr
 * 
 * This file is part of feedbackd.
 *
 *  feedbackd is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  feedbackd is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with feedbackd; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/**
 * @file feedbackd-agent.c
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>
#include <string.h>
#include <dlfcn.h>

#ifdef HAVE_GETOPT_H
#ifdef HAVE_LIBGNUGETOPT
#undef __GNU_LIBRARY__
#endif
#include <getopt.h>
#endif

#include <limits.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>

#include <sys/select.h>
#include <sys/time.h>

#include <signal.h>
#include <errno.h>

#include <libxml/xmlmemory.h>
#include <libxml/parser.h>

#include "necp.h"
#include "necp_client.h"
#include "feedbackd-agent.h"
#include "log.h"

/**
 * Where to find plugin modules
 */
#ifndef MODULE_DIR
#define MODULE_DIR "/usr/lib/feedbackd-agent/"
#endif

/**
 * Default location of config file
 */
#ifndef FEEDBACKD_CONF
#define FEEDBACKD_CONF "/etc/feedbackd-agent.xml"
#endif

/**
 * Default log location
 */
#ifndef FEEDBACKD_LOG
#define FEEDBACKD_LOG "/var/log/feedbackd-agent.log"
#endif

/**
 * How many times to attempt reconnection. 0 = try forever
 */
#ifndef CONNECTION_RETRIES
#define CONNECTION_RETRIES 10
#endif

/**
 * Amount of time (since last data received) to consider connection
 * to master broken
 */
#ifndef MASTER_TIMEOUT
#define MASTER_TIMEOUT 30
#endif



/* ANSI fun. */
extern int strncasecmp(const char *s1, const char *s2, size_t n);
struct in_addr server_address;
int setenv(const char *name, const char *value, int overwrite);

static char *moduledir;
static unsigned int connection_retries;
static int master_timeout;

static int configure(const char *configfile, int logging);

const struct option options[] = {
	{"conffile",    1, 0, 'c'},
	{"test-config", 1, 0, 't'},
	{"dont-fork",   0, 0, 'd'},
	{"help",        0, 0, 'h'},
	{"usage",       0, 0, 'u'},
	{"version",     0, 0, 'V'},
	{0,             0, 0, 0 }
};

int main(int argc, char **argv)
{
	char *conffile;
	int c, x, dofork = 1, testconfig = 0;

	/* set defaults */
	conffile = FEEDBACKD_CONF;
	moduledir = MODULE_DIR;
	connection_retries = 0;
	master_timeout = MASTER_TIMEOUT;

	while ((c = getopt_long(argc, argv, "c:tdVhu?", options, NULL))
	        != EOF) {

		switch(c) {
		case 'c':
			conffile = optarg;
			break;
		case 't':
			testconfig = 1;
			break;
		case 'd':
			dofork = 0;
			break;
		case 'V':
			printf("%s %s\n", PACKAGE_NAME, PACKAGE_VERSION);
			return 0;
		case '?':
		case 'h':
		case 'u':
		default:
			print_usage(argv[0]);
			return -1;
		}
	}

	
	log_printf(LOG_INFO, "%s starting", argv[0]);
	log_printf(LOG_INFO, "Using config file %s", conffile);

	/* random numbers requrired for necp request ids */
	srand((unsigned int)getpid());

	list_init(&services);

	/* read configuration */
	if (configure(conffile, !testconfig)) {
		log_printf(LOG_CRIT, "Configuration failed. Exiting");
		return -1;
	}

	if (testconfig) {
		log_setlevel(LOG_VDEBUG);
		dump_config();
		log_printf(LOG_VDEBUG, "Configuration sucessful");
		return 0;
	}

	dump_config();

	running = 1;
	
	if (dofork && fork())
		return 0;

	x = run();

	log_stop();

	return x;
}

static int configure(const char *configfile, int logging)
{
	list conf_services;
	listitem *i;
	xmlNodePtr lognode = NULL, node;
	xmlDocPtr doc;
	char *end;
	unsigned long tmp;

	assert(configfile && strlen(configfile));

	doc = xmlParseFile(configfile);
	if (!doc) {
		log_printf(LOG_ERR, "Could not parse config file %s",
				configfile);
		return -1 ;
	}

	node = xmlDocGetRootElement(doc);
	if (!node) {
		log_printf(LOG_ERR, "Configuration file %s is empty",
		    configfile);
		xmlFreeDoc(doc);
		return -1;
	}

	if (xmlStrcmp(node->name, (const xmlChar *) "feedbackd-agent")) {
		log_printf(LOG_ERR, "Configuration file %s has an incorrect "
				"root node (found %s, expected %s)", node->name,
				"feedbackd-agent");
		xmlFreeDoc(doc);
		return -1;
	}

	/* when support for reconfiguration is added, clear existing services
	 * table here */
	list_init(&conf_services);

	node = node->children;
	while (node) {
		char *value;
		
		if (xmlIsBlankNode(node)) {
			node = node->next;
			continue;
		}
		if (!xmlStrcmp(node->name, (const xmlChar*)"director")) {
			struct hostent *host;
			value = xmlNodeListGetString(doc, node->children, 1);
			if (!value) {
				log_printf(LOG_ERR, "No director address given "
						"in <director> tag");
				xmlFreeDoc(doc);
				return -1;
			}
			if (!(host = gethostbyname(value))
				|| host->h_addrtype != AF_INET
				|| host->h_length != sizeof(struct in_addr)) {
				log_printf(LOG_ERR, "Could not resolve "
						"director address %s", value);
				xmlFreeDoc(doc);
				return -1;
			}
			
			memcpy(&server_address, host->h_addr_list[0],
				       sizeof(struct in_addr));
		} else if (!xmlStrcmp(node->name,
				      (const xmlChar*)"moduledir")) {
			value =  xmlNodeListGetString(doc, node->children, 1);
			if (!value) {
				log_printf(LOG_ERR, "No value given in "
				             "<moduledir> tag");
				xmlFreeDoc(doc);
				return -1;
			}
			if (!(moduledir = malloc(strlen(value) + 1))) {
				perror("malloc");
				xmlFreeDoc(doc);
				return -1;
			}
			strcpy(moduledir, value);
			
		} else if (!xmlStrcmp(node->name,
				      (const xmlChar*)"connection-retries")) {
			value =  xmlNodeListGetString(doc, node->children, 1);
			if (!value) {
				log_printf(LOG_ERR, "No value given in "
				             "<connection-retries> tag");
				xmlFreeDoc(doc);
				return -1;
			}

			errno = 0;
			tmp = strtoul(value, &end, 10);
			if (*end != '\0' || end == value || errno == ERANGE
		            || tmp > UINT_MAX) {
				log_printf(LOG_ERR, "Invalid value (%s) given in "
				             "connection-retries tag.",
					     value);
				xmlFreeDoc(doc);
				return -1;
			}
			connection_retries = (unsigned int)tmp;
		} else if (!xmlStrcmp(node->name,
				      (const xmlChar*)"master-timeout")) {
			value =  xmlNodeListGetString(doc, node->children, 1);
			if (!value) {
				log_printf(LOG_ERR, "No value given in "
				             "<master-timeout> tag");
				xmlFreeDoc(doc);
				return -1;
			}

			errno = 0;
			tmp = strtoul(value, &end, 10);
			if (*end != '\0' || end == value || errno == ERANGE
		            || tmp > UINT_MAX) {
				log_printf(LOG_ERR, "Invalid value (%s) given "
						"in master-timeout tag.",
						value);
				xmlFreeDoc(doc);
				return -1;
			}
			master_timeout = (unsigned int)tmp;
		} else if (!xmlStrcmp(node->name, (const xmlChar*)"log")) {
			/* do log configuration last */
			lognode = node;
		} else if (!xmlStrcmp(node->name, (const xmlChar*)"service")) {
			/* service nodes are added to a list to be configured
			 * after all global directives have been processed */
			if (list_add(&conf_services, node)) {
				log_printf(LOG_ERR, "Error adding to list");
				return -1;
			}
		}
		node = node->next;
	}

	/* configure the services */
	list_init(&services);
	
	i = conf_services.head;

	while (i) {
		listitem *tmp;
		struct service *service = NULL;
		node = (xmlNodePtr)i->item;
		if ((service = parseService(node))) {
			if (list_add(&services, service)) {
				log_printf(LOG_ERR, "Error adding to service "
						"list.");
				return -1;
			}
		} else {
			log_printf(LOG_ERR, "Error in service definition in "
			             "config file %s", configfile);
			xmlFreeDoc(doc);
			return -1;
		}
		tmp = i;
		i = i->next;
		free(tmp);
	}


	/* now change the log over to the conf definition */
	if (logging && log_configure(lognode)) {
		log_printf(LOG_ERR, "Could not configure logging");
		return -1;
	}
		

	xmlFreeDoc(doc);
	return 0;

}

		
struct service *parseService(xmlNodePtr node)
{
	char *value, *tmpserv = 0;
	struct protoent *proto = 0;
	struct service *service;
	struct servent *serv;
	
	
	if (!(service = (struct service *)malloc(sizeof(struct service)))) {
		perror("malloc");
		return NULL;
	}
	memset(service, 0, sizeof(*service));
	
	if ((value = xmlGetProp(node, (const xmlChar *)"name"))) {
		strncpy(service->name, value, MAX_SERVICE_NAME_LEN - 1);
	} else {
		log_printf(LOG_ERR, "No name attribute for service");
		free(service);
		return NULL;
	}

	node = node->children;
	while (node) {
		if (xmlIsBlankNode(node)) {
			node = node->next;
			continue;
		}
		
		if (!xmlStrcmp(node->name, (const xmlChar*)"port")) {
			value = xmlNodeListGetString(node->doc,
					             node->children, 1);
			if (!value) {
				log_printf(LOG_ERR, "No port given in <port> "
						"tag, service %s",
						service->name);
				free(service);
				return NULL;
			}

			tmpserv = value;
			
		} else if (!xmlStrcmp(node->name,
		                     (const xmlChar *)"protocol")) {
			value = xmlNodeListGetString(node->doc,
			                             node->children, 1);
			if (!value) {
				log_printf(LOG_ERR, "No protocol given in "
						"<protocol> tag, service %s",
						service->name);
				free(service);
			}

			if (!(proto = getprotobyname(value))) {
				log_printf(LOG_ERR, "Invalid protocol (%s) "
						"given for service %s",
						value, service->name);
				free(service);
				return NULL;
			}
			service->protocol = proto->p_proto;

		} else if (!xmlStrcmp(node->name,
		                     (const xmlChar *)"forwarding")) {
			value = xmlNodeListGetString(node->doc,
			                             node->children, 1);
			if (!value) {
				log_printf(LOG_ERR, "No forwarding type given "
						"in <forwarding> tag, "
						"service %s",
						service->name);
				free(service);
				return NULL;
			}
			if (!(service->forwarding = get_forwarding(value))) {
				log_printf(LOG_ERR, "Invalid forwarding string "
						"(%s) given for service %s",
						value, service->name);
				free(service);
				return NULL;
			}

		} else if (!xmlStrcmp(node->name,
		                     (const xmlChar *)"monitor")) {
			const char *error;
			void *handle;
			char *modpath;
			void *(*cfgsym)(xmlNodePtr);
			void (*logsym)(void (loglevel, const char *, ...));

			if (!(value = xmlGetProp(node,
			             (const xmlChar *)"plugin"))) {
				log_printf(LOG_ERR, "No plugin attribute given "
						"for monitor tag in service "
						"%s.",
						service->name);
				free(service);
				return NULL;
			}
				
			modpath = malloc(strlen(moduledir) +
					 strlen(value) + 2);
			strcpy(modpath, moduledir);
			strcat(modpath, "/");
			strcat(modpath, value);

			if (!(handle = dlopen(modpath, RTLD_LAZY))) {
				log_printf(LOG_ERR, "Monitor plugin %s "
						"(defined in service %s) could "
						"not be loaded: %s",
						modpath, service->name,
						dlerror());
				free(modpath);
				free(service);
				return NULL;
			}
			

			/* Set the loghandle if the loghandle()
			 * function exists */
			logsym = dlsym(handle, "loghandle");
			if ((error = dlerror())) {
				log_printf(LOG_INFO, "No loghandle() symbol "
						"found in plugin %s, bypassing "
						"logging.", value);
			} else {
				(*logsym)(log_printf);
			}
			
			/* pass xml node to plugin if config() method
			 * exists */
			cfgsym = dlsym(handle, "config");
			if ((error = dlerror())) {
				log_printf(LOG_INFO, "No config() symbol found "
						"in plugin %s, bypassing "
						"configuration.", value);
				service->monitor.config = 0;
			} else {
				/* create a duplicate pointer */
				xmlNodePtr confignode =
					(xmlNodePtr)malloc(sizeof(xmlNode));
				memcpy(confignode, node, sizeof(xmlNode));
				service->monitor.config =
					(*cfgsym)(confignode);
				if (!service->monitor.config) {
					log_printf(LOG_ERR, "Configuration of "
							"plugin %s failed.",
							value);
					(void *)dlclose(handle);
					free(modpath);
					free(service);
					return NULL;
				}
				free(confignode);
			}

			service->monitor.destroy = dlsym(handle, "destroy");
			if ((error = dlerror())) {
				log_printf(LOG_WARN, "No destroy() symbol "
						"found in plugin %s.", value);
				service->monitor.destroy = 0;
			}
						
				
			/* set the dlsym property to the
			 * get_load function of the .so file
			 */
			service->monitor.get_health =
				dlsym(handle, "get_health");

			if ((error = dlerror())) {
				log_printf(LOG_ERR,
				    "Could not locate get_health() symbol in "
				    "%s", value);
				(void *)dlclose(handle);
				free(modpath);
				free(service);
				return NULL;
			}
				
			free(modpath);

		}
		node = node->next;
	}
	/* check service */
	if (!service->name || !strlen(service->name)) {
		log_printf(LOG_ERR, "No name given for service");
		free(service);
		return NULL;
	}

	if (!service->protocol) {
		log_printf(LOG_ERR, "No protocol given for service %s",
		             service->name);
		free(service);
		return NULL;
	}
	
	/* only look for port if the protocol is tcp or udp */
	if (service->protocol == IPPROTO_TCP ||
	    service->protocol == IPPROTO_UDP) {
		long port;
		char *end;
		
		if (!tmpserv) {
			log_printf(LOG_ERR, "No port given for service %s",
			     	service->name);
				free(service);
			return NULL;
		}

		errno = 0;
		port = strtol(tmpserv, &end, 10);
		
		if (*end == '\0' && end != tmpserv) {
			/* it's a number */
			if (errno == ERANGE || port < 0 || port > 0xffff) {
				log_printf(LOG_ERR, "Invalid port (%s) given "
						"for service %s",
					     tmpserv, service->name);
				free(service);
				return NULL;
			}
			service->port = htons(port);
		} else {
			
			if (!(serv = getservbyname(tmpserv, proto->p_name))) {
				log_printf(LOG_ERR, "Invalid port (%s) given "
						"for service %s",
						tmpserv, service->name);
				free(service);
				return NULL;
			}
			service->port = serv->s_port;
		}
	} else {
		log_printf(LOG_WARN, "Service has a protocol of %s, which is "
				"not supported by LVS", service->protocol);
		service->port = 0;
	}

	if (!service->forwarding) {
		log_printf(LOG_ERR, "No forwarding type given for service %s",
		             service->name);
		free(service);
		return NULL;
	}
	if (!service->monitor.get_health) {
		log_printf(LOG_ERR, "No monitor plugin given for service %s",
		             service->name);
		free(service);
		return NULL;
	}
	return service;
}


void dump_config()
{
	struct service *service;
	listitem *i = services.head;

	log_printf(LOG_DEBUG, "-- Configuration dump:");

	log_printf(LOG_DEBUG, "Director address: %s",
			inet_ntoa(server_address));
	log_printf(LOG_DEBUG, "Module Dir: %s", moduledir);
	log_printf(LOG_DEBUG, "Connection Retries: %d", connection_retries);

	while (i) {
		service = (struct service *)i->item;
		log_printf(LOG_DEBUG, "Service %s: "
		       "protocol: %3d "
		       "port: %5d "
		       "forwarding: %1d",
		       service->name,
		       service->protocol,
		       ntohs(service->port),
		       service->forwarding);
		i = i->next;
	}

	log_printf(LOG_DEBUG, "-- End of configuration dump");
		    
}

necp_ftype get_forwarding(const char *forwarding)
{

	assert(forwarding && strlen(forwarding));
	
	if (!strncasecmp(forwarding, "DR", 2)) {
		return NECP_FT_DROUTE;
	} else if (!strncasecmp(forwarding, "TUN", 3)) {
		return NECP_FT_TUNNEL;
	} else if (!strncasecmp(forwarding, "NAT", 3)) {
		return NECP_FT_MASQ;
	}

	return 0;

}

int run()
{
	int sd;
	unsigned int retries = 0;
	struct sockaddr_in addr;
	fd_set rfdset, xfdset;
	listitem *item;
	struct service *service;

connect:
	if (!running)
		return -1;
	
	/* create a socket */
	if ((sd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		perror("socket");
		if (connection_retries && retries++ == connection_retries) {
			log_printf(LOG_ERR, "Maximum number of connection "
					"attempts (%d) reached, giving up",
					connection_retries);
			return -1;
		}
		/* Check the signal issues here. --ratz */
		sleep(60);
		goto connect;
	}

	/* create address structure from server_address */
	memset ((char *)&addr, 0, sizeof(addr));
	addr.sin_family	= AF_INET;
	addr.sin_port	= htons(NECP_PORT);
	memcpy(&(addr.sin_addr), &server_address, sizeof(server_address));

	/* connect to server */
	if (connect(sd, (struct sockaddr *)&addr, sizeof(addr))) {
		perror("connect");
		close(sd);
		if (connection_retries && retries++ == connection_retries) {
			log_printf(LOG_ERR, "Maximum number of connection "
					"attempts (%d) reached, giving up",
					connection_retries);
			return -1;
		}
		sleep(60);
		goto connect;
	}

	/* we're connected now */
	retries = 0;
	
	/* send INIT packet */
	necp_init(sd);
	
	/* catch interrupt signals and call the interrupt() function,
	 * and ignore SIGPIPE errors (writing to a broken socket) */
	if (signal(SIGINT, interrupt) == SIG_ERR ||
	    signal(SIGPIPE, SIG_IGN) == SIG_ERR) {
		perror("signal");
		return -1;
	}
	
	/* flag to stop the client process, set by a SIGINT */
	stop_service = 0;
		
	while (running) {	
		int n;
		struct timeval timeout = {master_timeout, 0};

		FD_ZERO(&rfdset);
		FD_SET(sd, &rfdset);
		FD_ZERO(&xfdset);
		FD_SET(sd, &xfdset);
	
		/* wait for data on the socket */
		n = select(sd+1, &rfdset, NULL, &xfdset, &timeout);

		if (n < 0) {
			if (errno != EINTR) {
				perror("select");
				return -1;
			}
			/* check to see if an interrupt has occurred since
			 * the last iteration */
			if (stop_service) {
				log_printf(LOG_INFO, "Caught SIGINT, stopping "
						"services");
				necp_stop(sd);
			}
		} else if (n == 0) {
			log_printf(LOG_WARN, "Master timed out, "
					"attempting reconnection");
			close(sd);
			goto connect;
		} else {
			if (FD_ISSET(sd, &rfdset)) {
				int ret = data_received(sd);
				if (ret == -2) {
					/* recoverable problem - close socket
					 * and reconnect */
					log_printf(LOG_WARN, "Recoverable "
							"error, attempting "
							"reconnection");
					close(sd);
					goto connect;
				} else if (ret == -1) {
					/* crud. */
					return -1;
				}
			}
		}


	}

	/* change the signal handler back to the default */
	signal(SIGINT, SIG_DFL);
	signal(SIGPIPE, SIG_DFL);

	/* clean up monitors */
	item = services.head;
	while (item) {
		service = (struct service *)item->item;
		if (service->monitor.destroy) {
			log_printf(LOG_DEBUG, "Removing monitor plugin for "
			               "service %s", service->name);
			(*service->monitor.destroy)
				(service->monitor.config);
		}
		item = item->next;
	}

	log_printf(LOG_INFO, "Shutting down");

	return 0;
}

void interrupt(int signum)
{
	if (stop_service && signum == SIGINT)
		running = 0;
	else
		stop_service = 1;
}

void print_usage(const char *progname)
{
	fprintf(stderr,
"Usage: %s [-c|--conffile FILE] [-f|--dont-fork]"
"           [-t|--test-config]\n"
"\n"
"Options:\n"
"\n"
"    --conffile FILE:    Use configuration file FILE\n"
"                         (instead of default: %s)\n"
"\n"
"    --test-config:      Test configuration data, then exit.\n"
"\n"
"    --dont-fork:        don't run in the background. Useful with -f -\n"
"\n"
"    --version:          print version information and exit\n"
, progname, FEEDBACKD_CONF);
}
