/*
 * feedbackd - dynamic feedback system for LVS
 * Copyright (C) 2002 Jeremy Kerr
 * 
 * This file is part of feedbackd.
 *
 *  feedbackd is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  feedbackd is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with feedbackd; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/**
 * @file perl.c
 *
 * Monitor plugin to interface to a perl script.
 * This plugin requires a single configuration tag:
 *  <file>/path/to/perlscript.pl</file>
 * Which specifies a per script to parse when then plugin is initialised.
 * When the load value is required, a get_load() sub is called in the
 * perl file, which should return a SCALAR, between 0 and 100, representing
 * the system health (100 is perfect health, 0 is unable to serve requests).
 *
 * The perl script is only parsed once, but the get_health sub is called
 * whenever necessary.
 */

#include <errno.h>

#include <EXTERN.h>
#include <perl.h>
#include <XSUB.h>

#include <libxml/tree.h>

#include "plugin.h"

#define ERRSIZE 1024

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifndef call_pv
#define call_pv perl_call_pv
#endif

#ifdef PERL_MULTIPLICITY

typedef struct {
	char *file;
	PerlInterpreter *perl;
} perl_config;

#else

typedef struct {
	char *file;
} perl_config;

PerlInterpreter *perl;

/* super ugly hack flag. if this is 1, reparse config->file on
 * get_health() */
unsigned char perl_superuglyhack = 0;


#endif /* PERL_MULTIPLICITY */

char *perlargs[3];


void (*lh)(loglevel, const char *, ...) = NULL;

void _init()
{
	lh = NULL;
}

void loghandle(void (*logh)(loglevel, const char *, ...))
{
	lh = logh;
}

void *config(xmlNodePtr node)
{
	char *perlfile = NULL;
	struct stat statbuf;
	perl_config *cfg = 0;
	char err[ERRSIZE];
	node = node->children;

	while (node) {
		if (xmlIsBlankNode(node)) {
			node = node->next;
			continue;
		}
		if (!xmlStrcmp(node->name, (const xmlChar*)"file")) {
			if (!(perlfile = xmlNodeListGetString(node->doc,
					             node->children, 1))) {
				if (lh)
					(*lh)(LOG_ERR, "[PLUGIN perl] no "
						         "perl script "
							 "specified");
				return NULL;
			}
		}
		node = node->next;
	}
	if (!perlfile || !strlen(perlfile)) {
		if (lh) {
			(*lh)(LOG_ERR, "[PLUGIN perl] no perl script "
			                  "specified");
			return NULL;
		}
	}
	/* stat to make sure file exists */
	if (stat(perlfile, &statbuf)) {
		if (lh) {
			(*lh)(LOG_ERR, "[PLUGIN perl] could not read perl "
			                 "script %s: %s", perlfile,
					 strerror_r(errno,err,ERRSIZE));
			return NULL;
		}
	}

	cfg = (perl_config *)malloc(sizeof(perl_config));

	if (!cfg) {
		if (lh)
			(*lh)(LOG_ERR, "[PLUGIN perl] malloc: %s",
			               strerror_r(errno,err,ERRSIZE));
		return NULL;
	}
	
	if (lh)
		(*lh)(LOG_INFO, "[PLUGIN perl] using perl script %s",
		      perlfile);

	
	perlargs[0] = "feedbackd-agent";
	perlargs[1] = perlfile;
	perlargs[2] = NULL;

	cfg->file = (char *)malloc(strlen(perlfile) + 1);

	if (!cfg->file) {
		if (lh)
			(*lh)(LOG_ERR, "[PLUGIN perl] malloc: %s",
			               strerror_r(errno,err,ERRSIZE));
		return NULL;
	}
	strcpy(cfg->file, perlfile);

#ifdef PERL_MULTIPLICITY
	cfg->perl = perl_alloc();
	PERL_SET_CONTEXT(cfg->perl);
	perl_construct(cfg->perl);
	perl_parse(cfg->perl, NULL, 2, perlargs, (char **)NULL);
#else

	if (!perl_superuglyhack) {
		if (perl) {
			/* we've already set up an interpreter */
			perl_free(perl);
			perl_superuglyhack = 1;
			if (lh)
			(*lh)(LOG_WARN, "[PLUGIN perl] Multiple perl plugins "
			                "defined, but the perl interpreter "
					"does not support multiplicity. "
					"Continuing with less efficient "
					"monitor loading");
		} else {
			/* otherwise, this is the first perl monitor.
			 * Establish the single interpreter */
			perl = perl_alloc();
			perl_construct(perl);
			perl_parse(perl, NULL, 2, perlargs, (char **)NULL);
		}
	}
		
#endif
	
	return cfg;

}

int get_health(void *configdata)
{
	int health = 0, count;
	perl_config *cfg = (perl_config *)configdata;
	dSP;
	
#ifdef PERL_MULTIPLICITY
	PERL_SET_CONTEXT(cfg->perl);
#else
	if (perl_superuglyhack) {
		perl = perl_alloc();
		perl_construct(perl);
		perlargs[1] = cfg->file;
		perl_parse(perl, NULL, 2, perlargs, (char **)NULL);
	}
#endif

	if (lh)
		(*lh)(LOG_VDEBUG, "[PLUGIN perl] measuring health with "
		                  "script %s", cfg->file);
	
	ENTER;
	SAVETMPS;

	PUSHMARK(SP);
	PUTBACK;

	
	count = call_pv("get_health", G_NOARGS | G_SCALAR);
	SPAGAIN;
	
	if (count == 1)
		health = POPi;
	
	PUTBACK;
	FREETMPS;
	LEAVE;

#ifndef PERL_MULTIPLICITY
	if (perl_superuglyhack) {
		perl_destruct(perl);
		perl_free(perl);
	}
#endif
	
	return health;
}


int destroy(void *configdata)
{
	perl_config *cfg = (perl_config *)configdata;
#ifdef PERL_MULTIPLICITY
	PERL_SET_CONTEXT(cfg->perl);
	perl_destruct(cfg->perl);
	perl_free(cfg->perl);
#else
	if (!perl_superuglyhack) {
		perl_destruct(perl);
		perl_free(perl);
	}
		
	free(cfg->file);
#endif
	return 0;
}
	

void _fini()
{
}
